/*
 * yasapi.h
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef __YASAPI_H__ // [
#define __YASAPI_H__

#include <stdlib.h>
#include <stdio.h>
#include <winamp/out.h>
#include <yasapi_param.h>
#include <yasapi_api.h>

#ifdef __cplusplus // [
extern "C" {
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // [
typedef struct _Trace Trace;
#endif // ]
typedef enum _EEngine EEngine;
typedef const struct _IRingVtbl IRingVtbl;
typedef struct _Ring Ring;
typedef struct _Engine Engine;

#if defined (YASAPI_DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
struct _Trace {
  int bConsole;
};

int TraceCreate(Trace *pTrace, Out_Module *pOut);
void TraceDestroy(Trace *pTrace);
#endif // ]

///////////////////////////////////////////////////////////////////////////////
struct _IRingVtbl {
  struct {
//#define YASAPI_SOURCE_CONVERT_SIZE
#if defined (YASAPI_SOURCE_CONVERT_SIZE) // [
    DWORD (*ConvertSize)(Ring *pRing, DWORD dwSize);
#endif // ]
    DWORD (*Size)(Ring *pRing);
    DWORD (*Written)(Ring *pRing);
    DWORD (*Available)(Ring *pRing);
  } source;

  struct {
    DWORD (*ConvertSize)(Ring *pRing, DWORD dwSize);
    DWORD (*Size)(Ring *pRing);
    DWORD (*Written)(Ring *pRing);
//#define YASAPI_TARGET_AVAILABLE
#if defined (YASAPI_TARGET_AVAILABLE) // [
    DWORD (*Available)(Ring *pRing);
#endif // ]
  } target;

  //void *(*Malloc)(Ring *pRing, DWORD dwSize);
  BYTE *(*CopyToRing)(Ring *pRing, BYTE *dst, const BYTE *src, DWORD dwSize);
  BYTE *(*CopyFromRing)(Ring *pRing, BYTE *dst, const BYTE *src, DWORD dwSize);
};

extern IRingVtbl gcIRingPlainVtbl;
extern IRingVtbl gcIRingMono2StereoVtbl;

struct _Ring {
  IRingVtbl *lpVtbl;
  Engine *pEngine;
  BYTE *_buf;
  BYTE *_rp;
  BYTE *_wp;
  BYTE *_mp;
  DWORD _dwWritten;
  DWORD _dwAvailable;
};

int RingCreate(Ring *pRing, IRingVtbl *lpVtbl, Engine *pEngine, DWORD dwSize);
void RingDestroy(Ring *pRing);

void RingReset(Ring *pRing);
int RingWrite(Ring *pRing, BYTE *buf, DWORD dwSize);
int RingRead(Ring *pRing, BYTE *buf, DWORD dwSize);

///////////////////////////////////////////////////////////////////////////////
enum _ECommand {
  eCommandNull,
  eCommandExit,
  eCommandStart,
  eCommandFlush,
  eCommandKillDevice,
#if defined (YASAPI_OPTION_DISCONNECT) // [
  eCommandKillPauseDisconnected,
  eCommandKillPauseConnected,
#else // ] [
  eCommandKillPause,
#endif // ]
  ////////////////
  eCommandMax,
};

enum _EEngine {
  eEngineNull,
  eEngineCriticalSection,
  eEngineTimerEvent,
  eEngineTimer,
  eEngineEventProducer,
  eEngineEventConsumer,
  eEngineThread,
  eEngineThreadRunning,
  ////////////////
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  eEngineDeviceEnumerator,
  eEngineDevice,
#endif // ]
  eEngineAudioClientInititial,
  ////////////////
  eEngineRing,
#if defined (YASAPI_OPTION_DISCONNECT) // [
  eEnginePauseDisconnected,
#else // ] [
  eEnginePause,
#endif // ]
  eEngineAudioClient,
#if defined (YASAPI_OPTION_DISCONNECT) \
    && defined (YASAPI_PAUSE_CONNECTED2) // [
  eEnginePauseConnected2,
#endif // ]
#if defined (YASAPI_CLOCK) // [
  eEngineAudioClock,
#endif // ]
  eEngineAudioRenderClient,
#if defined (YASAPI_OPTION_DISCONNECT) \
    && ! defined (YASAPI_PAUSE_CONNECTED2) // [
  eEnginePauseConnected,
#endif // ]
  eEngineAudioClientStarted,
  eEngineTimerStarted,
  eEngineDraining,
  eEngineEot,
  ////////////////
  eEngineMax,
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  eEngineBase=eEngineDevice,
#else // ] [
  eEngineBase=eEngineThreadRunning,
#endif // ]
  eEngineStopped=eEngineAudioRenderClient,
  eEngineStarted=eEngineTimerStarted,
};

struct _Engine {
  EEngine eState;
  Player *pPlayer;
  CRITICAL_SECTION cs;

  struct {
    UINT32 uNumTracks;
    UINT32 uNumFramesMax;
  } stats;

  struct {
    HANDLE hEvent;
    HANDLE hTimer;
  } timer;

  struct {
    ECommand eCommand;
    HANDLE hProducer;     // When we're the producer we're waiting
                          // on this event.
    HANDLE hConsumer;     // When we're the consumer we're waiting
                          // on this event.
    HANDLE hThread;
  } thread;

#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  IMMDeviceEnumerator *pDeviceEnumerator;
#endif // ]
  IMMDevice *pDevice;

  struct {
    OptionsCommon common;
    OptionsDevice device;
  } options;

  struct {
    IRingVtbl *lpRingVtbl;
    const char *pFriendlyName;
    int bDoubleBuffered;
    AUDCLNT_SHAREMODE eShareMode;
    DWORD dwStreamFlags;
    REFERENCE_TIME hnsBufferDuration;
    LARGE_INTEGER DueTime;
    LONG lPeriod;
    UINT32 uNumFramesDevice;
    UINT32 uNumFramesRing;

    struct {
#if ! defined (YASAPI_OPTION_BITDEPTH) // [
      WORD nBlockAlign;
#endif // ]
      DWORD dwPlaybackBegin;
    } source;

#if defined (YASAPI_HISTORY) // [
    double msHistory;
#endif // ]
  } params;

  struct {
    double xWritten;
    double xPadding;
  } monitor;

#if defined (YASAPI_OPTION_BITDEPTH) // [
  struct {
    WAVEFORMATEXTENSIBLE source;
    WAVEFORMATEXTENSIBLE target;
  } wfx;
#else // ] [
  WAVEFORMATEXTENSIBLE wfx;
#endif // ]
  Ring ring;
  IAudioClient *pAudioClient;
#if defined (YASAPI_CLOCK) // [
  IAudioClock *pAudioClock;
#endif // ]
  IAudioRenderClient *pAudioRenderClient;
};

int EngineCreate(Engine *pEngine, Player *pPlayer, int samplerate,
    int numchannels, int bitspersamp);
int EngineCreateAudioClient(Engine *pEngine, const EEngine eState);
void EngineKill(Engine *pEngine, const EEngine eState);
void EngineDestroy(Engine *pEngine);
void EngineClose(Engine *pEngine);

int EngineNumFramesPadding(Engine *pEngine, UINT32 *puNumFramesPadding);
int EnginePlayingStart(Engine *pEngine);
int EnginePause(Engine *pEngine, int bPause);
int EngineIsPlaying(Engine *pEngine);
int EngineAvailable(Engine *pEngine);
int EngineOutputTimeExt(Engine *pEngine);
int EngineWrittenTime(Engine *pEngine);
int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSize);

void EngineSend(Engine *pEngine, const ECommand eCommand, int bLock);

int EngineStart(Engine *pEngine, int bAudioClient, int bTimer);
int EnginePlay(Engine *pEngine);
void EngineFlush(Engine *pEngine);

DWORD WINAPI EngineThreadProc(LPVOID lpParameter);

#if defined (YASAPI_DEBUG) // [
const char *EngineCommand(Engine *pEngine);
const char *EngineState(Engine *pEngine);
#endif // ]

///////////////////////////////////////////////////////////////////////////////
struct _Player {
  Out_Module *pOut;
  wchar_t ini[MAX_PATH];
#if defined (YASAPI_DEBUG) // [
  Trace trace;
#endif // ]
  struct {
    UINT32 uNumTracks;
    UINT32 uMsLatencyMax;
  } stats;

  Wasapi wasapi;
  Engine engine;
  Gtk gtk;
};

int PlayerCreate(Player *pPlayer, Out_Module *pOut);
void PlayerDestroy(Player *pPlayer);

int PlayerPropertiesPath(Player *pPlayer);
int PlayerOpen(Player *pPlayer, int samplerate, int numchannels,
    int bitspersamp);
void PlayerClose(Player *pPlayer);

Wasapi *PlayerWasapi(Player *pPlayer);

void PlayerFlush(Player *pPlayer);
int PlayerPause(Player *pPlayer, int bPause);
int PlayerIsPlaying(Player *pPlayer);
int PlayerAvailable(Player *pPlayer);
int PlayerOutputTime(Player *pPlayer);
int PlayerWrittenTime(Player *pPlayer);
int PlayerWrite(Player *pPlayer, BYTE *buf, int len);

int PlayerShow(Player *pPlayer, Dialog *pDialog, HWND hWndParent);

#ifdef __cplusplus // [
}
#endif // ]
#endif // __YASAPI_H__ ]
