/*
 * yasapi_config.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
const char CONFIG_WIDGET[]="ConfigWidget";

///////////////////////////////////////////////////////////////////////////////
char *NavigatePlayer(Dialog *pDialog, int offs)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;

  return ((char *)pPlayer)+offs;
}

char *NavigateDeviceSelectedConfig(Dialog *pDialog, int offs)
{
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  WasapiDevice *pWasapiDevice=pWasapi->selected.pDeviceConfig;

  return ((char *)pWasapiDevice)+offs;
}

char *NavigateOptionsCommonConfig(Dialog *pDialog, int offs)
{
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  OptionsCommon *pOptionsCommon=&pWasapi->options.common.config;

  return ((char *)pOptionsCommon)+offs;
}

char *NavigateOptionsDeviceConfig(Dialog *pDialog, int offs)
{
  Wasapi *pWasapi=PlayerWasapi(pDialog->pGtk->pPlayer);
  WasapiDevice *pWasapiDevice=pWasapi->selected.pDeviceConfig;
  OptionsDevice *pOptionsDevice=&pWasapiDevice->pMember->config;

  return ((char *)pOptionsDevice)+offs;
}

char *NavigateEngine(Dialog *pDialog, int offs)
{
  Engine *pEngine=&pDialog->pGtk->pPlayer->engine;

  return eEngineNull<pEngine->eState?((char *)pEngine+offs):NULL;
}

///////////////////////////////////////////////////////////////////////////////
ConfigWidget *ConfigWidgetCreate(ClassCell *pCell, Dialog *pDialog,
    IConfigWidgetVtbl *lpVtbl, GtkWidget *widget)
{
  ConfigWidget *pWidget;

  if (0<=pCell->widget&&lpVtbl) {
    pWidget = (ConfigWidget *)((char *)&pDialog->config + pCell->widget);
    pWidget->lpVtbl=lpVtbl;
    pWidget->pDialog=pDialog;
    pWidget->pCell=pCell;
    pWidget->widget=widget;
    g_object_set_data(G_OBJECT(widget), CONFIG_WIDGET, pWidget);
    // It's not a good idea to call "set()" on our individual widget
    // because it may have side effects to other, not yet initialized
    // widgets.
    // That's the theory. In practice we need both.
    lpVtbl->set(pWidget);

#if ! defined (YASAPI_CLOCK) // [
    if (pWidget==&pDialog->config.common.latency)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
#if ! defined (YASAPI_OPTION_VOLUME) // [
    if (pWidget==&pDialog->config.common.volume)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
#if ! defined (YASAPI_OPTION_GAPLESS) // [
    if (pWidget==&pDialog->config.common.gapless)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
#if ! defined (YASAPI_OPTION_BITDEPTH) // [
    if (pWidget==&pDialog->config.yasapi.bitdepth)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
#if 0 && ! defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
    if (pWidget==&pDialog->config.wasapi.pro_audio)
      gtk_widget_set_sensitive (widget,0);
#endif // ]
  }
  else
    pWidget = NULL;

  if (pCell->tooltip)
    //gtk_widget_set_tooltip_text(widget, pCell->tooltip);
    gtk_widget_set_tooltip_markup (widget, pCell->tooltip);

  return pWidget;
}

static void ConfigWindowCreate(Dialog *pDialog)
{
  /////////////////////////////////////////////////////////////////////////////
  static ConfigComboBoxMap mode_map[]={
    { "Shared", eModeShared },
    { "Exclusive", eModeExclusive },
    { NULL, -1 }
  };

  static ConfigComboBoxMap strategy_map[]={
    { "Push", eStrategyPush },
    { "Pull", eStrategyPull },
    { NULL, -1 }
  };

  static const char attDevice[]="Choose a WASAPI enabled device.";

  static const char attMode[]=
    "Whether the plug-in should operate with WASAPI mode"
      " \"exclusive\" or \"shared\"."
    ;
  static const char attStrategy[]=
    "Whether the plug-in should operate with WASAPI strategy"
      " \"push\" or \"pull\"."
    ;

  static ClassPage general_rows = {
    { LABEL_STATIC(1,0,"Device:",attDevice),
        COMBO_BOX_DEVICE_WASAPI(1,0,general.device,attDevice) },
    { LABEL_STATIC(1,0,"Mode:",attMode), COMBO_BOX_DEVICE(1,0,
        general.mode,eMode,mode_map,attMode) },
    { LABEL_STATIC(1,0,"Strategy:",attStrategy), COMBO_BOX_DEVICE(1,0,
        general.strategy,eStrategy,strategy_map,attStrategy) },
    { { NULL } }
  };

  ////////////////
  static const char attLatency[]=
      "Whether YASAPI NT should optimize getoutputtime() and getwrittentime()"
      " by not using IAudioClock and always return zero and the latency"
      " resulting from YASAPI NT's buffer, respectively."
#if ! defined (YASAPI_CLOCK) // [
      " Not implemented yet. Always enabled, i.e. IAudioClock is not used."
#endif // ]
      ;
  static const char attMono2Stereo[]=
      "Duplicate the only channel from mono to a second channel for (pseudo)"
      " stereo (2.0)."
      ;
  static const char attVolume[]=
      "Enable volume control."
#if ! defined (YASAPI_OPTION_VOLUME) // [
      " Not implemented yet. Always disabled."
#endif // ]
      ;
  static const char attSurround[]="How 4, 6, and 8 channels shoud be"
      " interpreted: as surround, 5.1 surround, and 7.1 surround (enabled),"
      " or as quadrophonic, 5.1, and 7.1 (disabled)."
      ;
#if defined (YASAPI_OPTION_DISCONNECT) // [
  static const char attDisconnect[]="Wether YASAPI NT should"
      " release it's instances of IAudioClient, IAudioRenderClient, and"
      " IAudioClock during pause and re-initialize them when done."
      ;
#endif // ]
  static const char attGapless[]=
      "Enable gapless playback."
#if ! defined (YASAPI_OPTION_VOLUME) // [
      " Not implemented yet. Always disabled."
#endif // ]
      ;
  static const char attConfigPos[]=
    "Where the \"configuration\" dialog should be positioned."
    ;
  static const char attAboutPos[]=
    "Where the \"about\" dialog should be positioned."
    ;

  static ConfigComboBoxMap mono2stero_map[]={
    { "Off",eMono2StereoOff },
    { "Duplicate Plain",eMono2StereoPlain },
    { "Duplicate Mutiplied by 0.5",eMono2StereoHalf },
    { "Duplicate Mutiplied by 1.0/Sqrt(2.0)",eMono2StereoSqrt2Inverse },
    { NULL, -1 }
  };

  static ConfigComboBoxMap pos_map[]={
    { "Free Floating",eDialogPosFreeFloating },
    { "On Top of Preferences",eDialogPosOnTopPreferences },
    { "Always on Top",eDialogPosAlwaysOnTop },
    { NULL, -1 }
  };

  static ClassPage common_rows={
    { LABEL_STATIC(1,0,"Report Latency Instead of Time:",attLatency), 
        CHECK_BUTTON_COMMON(1,0,common.latency,bLatency,attLatency) },
    { LABEL_STATIC(1,0,"Play Mono as Stereo:",attMono2Stereo), 
        COMBO_BOX_COMMON(1,0,common.mono2stereo,eMono2Stereo,
            mono2stero_map,attMono2Stereo) },
    { LABEL_STATIC(1,0,"Volume Control:",attVolume), 
        CHECK_BUTTON_COMMON(1,0,common.volume,bVolume,attVolume) },
    { LABEL_STATIC(1,0,"Surround:",attSurround), 
        CHECK_BUTTON_COMMON(1,0,common.surround,bSurround,attSurround) },
#if defined (YASAPI_OPTION_DISCONNECT) // [
    { LABEL_STATIC(1,0,"Disconnect when Pause:",attDisconnect), 
        CHECK_BUTTON_COMMON(1,0,common.disconnect,bDisconnect,attDisconnect) },
#endif // ]
    { LABEL_STATIC(1,0,"Gapless Playback:",attGapless), 
        CHECK_BUTTON_COMMON(1,0,common.gapless,bGapless,attGapless) },
    { LABEL_STATIC(1,0,"Configuration Dialog Position:",attConfigPos), 
        COMBO_BOX_COMMON(1,0,common.pos.config,pos.eConfig,pos_map,
            attConfigPos) },
    { LABEL_STATIC(1,0,"About Dialog Position:",attAboutPos), 
        COMBO_BOX_COMMON(1,0,common.pos.about,pos.eAbout,pos_map,
            attAboutPos) },
    { { NULL } }
  };

  ////////////////
  static ConfigComboBoxMap period_map[]={
    { "Default", ePeriodDefault },
    { "Minimum", ePeriodMinimum },
    { NULL }
  };

  static const char attAutoConvertPCM[]=
      "Whether WASAPI should automatically convert to the desiered format."
      " Available only with mode \"shared\". Disabled otherwise."
      ;
  static const char attSRCDefaultQuality[]=
      "Whether (according to Microsoft) \"a sample rate converter with"
      " better quality than the default conversion but with a higher"
      " performance cost is used. This should be used if the audio is"
      " ultimately intended to be heard by humans as opposed to other"
      " scenarios such as pumping silence or populating a meter.\""
      ;
  static const char attTimerPeriodic[]=
    "Microsoft waitable timers are not strictly WASAPI objects. YASAPI NT"
    " uses one in order to implement strategy \"push\" and the option"
    " fits perfectly on this page. If enabled YASAPI NT starts the timer"
    " once initialized with an appropriate period matching a playback cycle."
    " Otherwise the timer ist re-started again with each playback cycle."
    ;
  static const char attDevicePeriod[]=
    //"Associated with what is choosen for WASAPI mode. According"
    "Associated with WASAPI mode. According"
    " to Microsoft \"[Default] specifies the [...] period for a"
    " shared-mode stream. [Minimum] specifies the [...] period for"
    " an exclusive-mode stream.\""
    ;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  static const char attProAudio[]=
    "Should at least be enabled with mode \"exclusive\" and"
    " strategy \"pull\"."
    ;
#endif // ]
  static const char attAlignmentDance[]=
    "Do the \"alignment dance\" when error AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED"
    " occurs with IAudioClient::Initialize()."
    ;

  static ClassPage wasapi_rows_frame={
    { LABEL_STATIC(1,0,"Mode:",attMode),LABEL_MAPPED_DEVICE(2,0,
        wasapi.mode,eMode,mode_map,attMode) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.auto_convert_pcm,bAutoConvertPCM,
            attAutoConvertPCM),
        LABEL_STATIC(1,0,"Auto Convert PCM",attAutoConvertPCM) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.src_default_quality,bSRCDefaultQuality,
            attSRCDefaultQuality),
        LABEL_STATIC(1,0,"SRC Default Quality",
            attSRCDefaultQuality) },
    { LABEL_STATIC(1,0,"Device Period:",attDevicePeriod),
        LABEL_MAPPED_DEVICE(2,0,wasapi.period,ePeriod,period_map,
            attDevicePeriod) },
    { LABEL_STATIC(1,0,"Strategy:",attStrategy),
        LABEL_MAPPED_DEVICE(2,0,wasapi.strategy,eStrategy,strategy_map,
            attStrategy) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.timer_periodic,bTimerPeriodic,
            attTimerPeriodic),
        LABEL_STATIC(1,0,"Periodic Timer",attTimerPeriodic) },
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
    { LABEL_STATIC(1,0,"Associate with \"Pro Audio\" Task:",
        attProAudio), CHECK_BUTTON_DEVICE(2,0,wasapi.pro_audio,bProAudio,
            attProAudio) },
#endif // ]
    { LABEL_STATIC(1,0,"Alignment Dance:",attAlignmentDance),
        CHECK_BUTTON_DEVICE(2,0,wasapi.alignment_dance,bAlignmentDance,
            attAlignmentDance) },
    { { NULL } }
  };

  static ClassPage wasapi_rows={
    { LABEL_DEVICE_WASAPI(1,0,wasapi.device) },
    { FRAME(wasapi_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ConfigComboBoxMap bitdepth_map[]={
    { "As Delivered by Winamp", eBitdepthDisabled },
    { "16 Bit", eBitdepth16 },
    { "24 Bit", eBitdepth24 },
    { NULL, -1 }
  };

  static ConfigScaleParameter ring_size_param={
    .min=2.0,
    .max=30.0,
    .step=0.1,
  };

  static ConfigScaleParameter latency_param={
    .min=0.0,
    .max=1000.0,
    .step=10.0,
  };

  static ConfigScaleParameter playback_begin_param={
    .min=1.0,
    .max=10.0,
    .step=0.1,
  };

  static const char attBitdepth[]=
    "Whether YASAPI NT should convert to the choosen format."
#if ! defined (YASAPI_OPTION_BITDEPTH) // [
    " Not implemented yet. Always set to \"as delivered by winamp\","
    " i.e. no conversion."
#endif // ]
    ;
  static const char attRingSize[]=
    "The size of YASAPI NT's buffer in terms of a packet's size"
    " as delivered by Winamp or the current input plug-in, respectively."
    ;
  static const char attChunkSize[]=
    "The size of an audio packet as delivered by Winamp or the current"
    " input plug-in, respectively, in order that YASAPI NT sends it to"
    " WASAPI."
    ;

  static ClassPage yasapi_rows_frame={
    { LABEL_STATIC(1,0,"Bit Depth:",attBitdepth),
        COMBO_BOX_DEVICE(1,0,yasapi.bitdepth,eBitdepth,bitdepth_map,
            attBitdepth) },
    { LABEL_STATIC(1,0,"Ring Buffer Size:",attRingSize),
        SCALE_DEVICE(2,0,yasapi.ring_size,xRingSize,&ring_size_param,
            attRingSize),
        LABEL_STATIC(3,0,"× Chunk Size",attChunkSize) },
    { LABEL_STATIC(1,0,"Max. Latency:",NULL),
        SCALE_DEVICE(2,0,yasapi.max_latency_assumed,xMaxLatencyAssumed,
            &latency_param,NULL),
            LABEL_STATIC(3,0,"ms (Educated Guess)",NULL) },
    { LABEL_STATIC(1,0,"Begin Playback:",NULL),
        SCALE_DEVICE(2,0,yasapi.playback_begin,xPlaybackBegin,
            &playback_begin_param,NULL),
        LABEL_STATIC(1,0,"×",NULL),
        LABEL_MAPPED_DEVICE(1,0,yasapi.period_playback_begin,ePeriod,
            period_map,attDevicePeriod),
        LABEL_STATIC(1,0,"Buffer Size",NULL) },
    { { NULL } }
  };

  static ClassPage yasapi_rows={
    { LABEL_DEVICE_WASAPI(1,0,yasapi.device) },
    { FRAME(yasapi_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ConfigLevelBarParameter param_monitor_written={ 0 };
  static ConfigLevelBarParameter param_monitor_padding={ 0 };

  static const char attWinampNumTracks[]=
    "The overall number of individual tracks played with the current Winamp"
    " session and any device."
    ;
  static const char attWinampNumGaplessSequences[]=
    "The overall number of gapless track sequences played with the current"
    " Winamp session and any device. A gapless track sequence may in turn"
    " consist of several individual tracks sharing the same format. Tracks"
    " of different format cannot played gaplessly"
    ;
  static const char attWinampMaxLatency[]=
    "The latency accumulated so far with the current Winamp session"
    " and any device."
    ;
  static const char attPadding[]=
    "Taken at the beginning of a playback cycle, the amount of not yet"
    " played sound cached in the WASAPI device buffer."
    " (Not defined in case of double buffering, i.e. exclusive pulled"
    " playback. Remains empty in this case.)"
    ;
  static const char attWritten[]=
    "Taken at the beginning of a playback cycle, the amount of not yet"
    " played sound cached in the YASAPI NT buffer."
    ;
  static const char attCurDevice[]=
    "The device in use."
    ;
  static const char attTracksEngine[]=
    "The number of tracks played within the current gapless sequence."
    ;
  static const char attFramesDevice[]=
    "The size of the buffer shared with the WASAPI device."
    ;
  static const char attFramesYasapi[]=
    "The size of the YASAPI NT buffer."
    ;
  static const char attLatencyGapless[]=
    "The latency accumulated so far with the current gapless track sequence."
    ;

  static ClassPage monitor_rows_frame={
    // {
    { LABEL_STATIC(1,0,"<b>Winamp Session</b>",NULL) },
    { LABEL_STATIC(1,0,"#Tracks:",attWinampNumTracks),
        LABEL_UINT32_PLAYER(1,0,
        "%u",monitor.player.num_tracks,stats.uNumTracks,attWinampNumTracks) },
    { LABEL_STATIC(1,0,"#Gapless Sequences:",attWinampNumGaplessSequences),
        LABEL_UINT32_PLAYER(1,0,
        "%u",monitor.player.num_gapless,stats.uNumTracks,
            attWinampNumGaplessSequences) },
    { LABEL_STATIC(1,0,"Max. Latency:",attWinampMaxLatency),
        LABEL_UINT32_PLAYER(1,0,
        "%u ms",monitor.player.ms_latency,stats.uMsLatencyMax,
            attWinampMaxLatency) },
    // }
    { LABEL_STATIC(1,0,"",NULL) },
    // {
    { LABEL_STATIC(1,0,"<b>Current Track"
        " / Current Gapless Track Sequence</b>",NULL) },
    { LABEL_STATIC(1,0,"Device:",attCurDevice),
        LABEL_ENGINE(1,0,monitor.device,params.pFriendlyName,attCurDevice) },
    { LABEL_STATIC(1,0,"#Tracks:",attTracksEngine), LABEL_UINT32_ENGINE(1,0,
        "%u",monitor.engine.num_tracks,stats.uNumTracks,attTracksEngine) },
    // {
    { LABEL_STATIC(1,0,"WASAPI Device Buffer",NULL) },
    { LABEL_STATIC(1,0,"Size:",attFramesDevice), LABEL_FRAMES_ENGINE(1,0,
        monitor.engine.num_frames_device,params.uNumFramesDevice,
            attFramesDevice) },
    { LABEL_STATIC(1,0,"Padding:",attPadding), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_padding,monitor.padding,monitor.xPadding,attPadding) },
    // }
    // {
    { LABEL_STATIC(1,0,"YASAPI NT Buffer",NULL) },
    { LABEL_STATIC(1,0,"Size:",attFramesYasapi), LABEL_FRAMES_ENGINE(1,0,
        monitor.engine.num_frames_ring,params.uNumFramesRing,
            attFramesYasapi) },
    { LABEL_STATIC(1,0,"Max. Latency:",attLatencyGapless),
        LABEL_FRAMES_ENGINE(1,0,monitor.engine.num_frames_latency,
            stats.uNumFramesMax,attLatencyGapless) },
    { LABEL_STATIC(1,0,"Padding:",attWritten), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_written,monitor.written,monitor.xWritten,attWritten) },
    // }
    // }
    { { NULL } }
  };

  static ClassPage monitor_rows={
    { FRAME(monitor_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ClassPage explore_rows_frame={
    { { NULL },
        LABEL_MATRIX_BITS(1,1,0),
        LABEL_MATRIX_BITS(1,1,1),
        LABEL_MATRIX_BITS(1,1,2) },
    { LABEL_MATRIX_FREQ(1,1,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,2) },
    { LABEL_MATRIX_FREQ(1,1,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,2) },
    { LABEL_MATRIX_FREQ(1,1,2),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,2) },
    { LABEL_MATRIX_FREQ(1,1,3),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,2) },
    { LABEL_MATRIX_FREQ(1,1,4),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,2) },
    { LABEL_MATRIX_FREQ(1,1,5),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,2) },
    { LABEL_MATRIX_FREQ(1,1,6),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,2) },
    { LABEL_MATRIX_FREQ(1,1,7),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,2) },
    { { NULL } }
  };

  static ClassPage explore_rows={
    { LABEL_DEVICE_WASAPI(1,0,explore.device) },
    { LABEL_STATIC(1,0,"Matrix of WASAPI Supported Integer Formats"
        " for Stereo (2.0)/Exclusive",NULL) },
    { FRAME(explore_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ClassNotebook notebook_pages={
    { LABEL_STATIC(1,0,"Common","Configure options independent of the"
        " choosen device."), PAGE(common_rows) },
    { LABEL_STATIC(1,0,"Wasapi","Configure options specific to the"
        " choosen device w.r.t Microsoft WASAPI."), PAGE(wasapi_rows) },
    { LABEL_STATIC(1,0,"Yasapi","Configure options specific to the"
        " choosen device w.r.t the YASAPI NT plug-in."), PAGE(yasapi_rows) },
    { LABEL_STATIC(1,0,"Monitor","Monitor running YASAPI NT. Nothing"
        " to configure."), PAGE(monitor_rows) },
    { LABEL_STATIC(1,0,"Explore","Explore capabilities of Microsoft"
        " WASAPI w.r.t the choosen device. Nothing to configure."),
        PAGE(explore_rows) },
    { { NULL } }
  };

  ////////////////
  static ClassPage rows={
    { PAGE(general_rows) },
    { NOTEBOOK(notebook_pages) },
    { BUTTON_BOX() },
    { { NULL } }
  };

  static ClassCell cell=PAGE(rows);

  /////////////////////////////////////////////////////////////////////////////
  GtkWidget *window,*widget;
  Expand e;

  // window.
  pDialog->window = window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  // box.
  e.h = FALSE;
  e.v = FALSE;
  widget = cell.lpVtbl->create (&cell,pDialog,&e);
  g_object_set (G_OBJECT (widget), "margin", ePanelSpace, NULL);

  if (e.h) {
    gtk_widget_set_hexpand (widget, TRUE);
    gtk_widget_set_halign (widget, GTK_ALIGN_FILL);
  }
  else
    gtk_widget_set_halign (widget, GTK_ALIGN_CENTER);

  if (e.v) {
    gtk_widget_set_vexpand (widget, TRUE);
    gtk_widget_set_valign (widget, GTK_ALIGN_FILL);
  }
  else
    gtk_widget_set_valign (widget, GTK_ALIGN_CENTER);

  gtk_container_add (GTK_CONTAINER (window), widget);
  DWRITELN("\"box\" added to \"window\".");

  // It's better to call "set()" on all of our widgets after all of them
  // have been initialized because some of them may have side effects on
  // others.
  pDialog->lpVtbl->set(pDialog);
}

static void ConfigApply(Dialog *pDialog, int moffs, const char *op)
{
#if 0 // [
  enum { eNumWidgets=&pDialog->config.explore.device
      - &pDialog->config.general.device
      + 1 };
#else // ] [
  enum {
    eWidgetSize=sizeof (ConfigWidget),
    eNumWidgetsGeneral=sizeof (pDialog->config.general)/eWidgetSize,
    eNumWidgetsCommon=sizeof (pDialog->config.common)/eWidgetSize,
    eNumWidgetsWasapi=sizeof (pDialog->config.wasapi)/eWidgetSize,
    eNumWidgetsYasapi=sizeof (pDialog->config.yasapi)/eWidgetSize,
    eNumWidgetsMonitor=sizeof (pDialog->config.monitor)/eWidgetSize,
    eNumWidgetsExplore=sizeof (pDialog->config.explore)/eWidgetSize,
    eNumWidgets
        =eNumWidgetsGeneral
        +eNumWidgetsCommon
        +eNumWidgetsWasapi
        +eNumWidgetsYasapi
        +eNumWidgetsMonitor
        +eNumWidgetsExplore,
  };
#endif // ]
  typedef void (method_t)(ConfigWidget *pWidget);

  ConfigWidget *pWidget=pDialog->config.widgets;
  ConfigWidget *pWidgetMax=pWidget + eNumWidgets;
  method_t *method;

//DVWRITELN("%s { ===================================================",op);
  while (pWidget<pWidgetMax) {
    if (pWidget->lpVtbl) {
      DWRITELN (pWidget->pCell->pId);
      method=*(method_t **)((char *)pWidget->lpVtbl + moffs);
      method(pWidget);
    }
    else
      DVWRITELN ("*** missing lpVtbl: %d ***",pWidget-pDialog->config.widgets);

    ++pWidget;
  }
//DVWRITELN("%s } ===================================================",op);
}

static void ConfigSet(Dialog *pDialog)
{
  ConfigApply(pDialog,offsetof(IConfigWidgetVtbl,set),"set");
}

static void ConfigGet(Dialog *pDialog)
{
  ConfigApply(pDialog,offsetof(IConfigWidgetVtbl,get),"get");
}

IDialogVtbl gIConfigVtbl={
#if defined (YASAPI_DEBUG) // [
#if defined (YASAPI_OPTIONS_DIALOG) // [
  .pIdA="Config",
#else // ] [
  .pId="Config",
#endif // ]
#endif // ]
#if defined (YASAPI_OPTIONS_DIALOG) // [
  .pIdW=L"Config",
#endif // ]
  .pTitle="Configure " YASAPI_TITLE,
  .nWidthPixels=eConfigWidthPixels,
  .WindowCreate=ConfigWindowCreate,
  .set=ConfigSet,
  .get=ConfigGet,
};
