/*
 * yasapi_thread.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
//#define YASAPI_THREAD_PLAYING

///////////////////////////////////////////////////////////////////////////////
#if ! defined (ERROR_INVALID_TASK_NAME) // [
#define ERROR_INVALID_TASK_NAME 1550L
#endif // ]

#if ! defined (ERROR_INVALID_TASK_INDEX) // [
#define ERROR_INVALID_TASK_INDEX 1551L
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef enum _EThreadLimits EThreadLimits;
typedef struct _Thread Thread;

///////////////////////////////////////////////////////////////////////////////
enum _EThreadLimits {
  eThreadCount=3,
};

struct _Thread {
  Engine *pEngine;
  HANDLE aHandles[eThreadCount];
#if defined (YASAPI_THREAD_PLAYING) // [
  int bPlaying;
#endif // ]
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  HANDLE hTask;
  DWORD dwTaskIndex;
#endif // ]
};

///////////////////////////////////////////////////////////////////////////////
static void ThreadCreate(Thread *pThread, Engine *pEngine)
{
  pThread->pEngine=pEngine;
  pThread->aHandles[0]=pEngine->thread.hProducer;
  pThread->aHandles[1]=pEngine->timer.hEvent;
  pThread->aHandles[2]=pEngine->timer.hTimer;
#if defined (YASAPI_THREAD_PLAYING) // [
  pThread->bPlaying=0;
#endif // ]
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  pThread->hTask=NULL;
  pThread->dwTaskIndex=0ul;
#endif // ]

  EnterCriticalSection(&pEngine->cs); // {
  ++pEngine->eState;
  DASSERT(eEngineThreadRunning==pEngine->eState,exit);
  DWRITELN(EngineState(pEngine));
  SetEvent(pEngine->thread.hConsumer);
#if defined (YASAPI_DEBUG) // [
exit:
#endif // ]
  return;
}

static void ThreadDestroy(Thread *pThread)
{
  Engine *pEngine=pThread->pEngine;

  DWRITELN(EngineState(pEngine));
  LeaveCriticalSection(&pEngine->cs); // }
  ExitThread(0ul);
}

static DWORD ThreadWait(Thread *pThread)
{
  Engine *pEngine=pThread->pEngine;
  DWORD dwWait;

  LeaveCriticalSection(&pEngine->cs); // {

  dwWait=WaitForMultipleObjects(
    eThreadCount,       // _In_       DWORD  nCount,
    pThread->aHandles,  // _In_ const HANDLE *lpHandles,
    FALSE,              // _In_       BOOL   bWaitAll,
    INFINITE            // _In_       DWORD  dwMilliseconds
  );

  EnterCriticalSection(&pEngine->cs); // }

  return dwWait;
}

static void ThreadKill(Thread *pThread, const EEngine eState)
{
  Engine *pEngine=pThread->pEngine;

#if defined (YASAPI_THREAD_PLAYING) // [
  pThread->bPlaying=0;
#endif // ]

  if (eEngineNull<eState)
    EngineKill(pEngine,eState);

#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  if (pEngine->eState<eEngineAudioClient&&pThread->hTask) {
    AvRevertMmThreadCharacteristics(pThread->hTask);
    pThread->hTask=NULL;
    pThread->dwTaskIndex=0ul;
  }
#endif // ]

  DWRITELN(EngineState(pEngine));
}

static int ThreadCommand(Thread *pThread)
{
  int code=-1;
  Engine *pEngine=pThread->pEngine;
  ECommand eCommand=pEngine->thread.eCommand;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  DWORD dwLastError;
#endif // ]

  switch (eCommand) {
  case eCommandStart:
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
    if (pEngine->options.device.bProAudio&&!pThread->hTask) {
      pThread->hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",
          &pThread->dwTaskIndex);

      if (!pThread->hTask) {
        dwLastError=GetLastError();

        switch (dwLastError) {
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_INVALID_TASK_INDEX);
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_INVALID_TASK_NAME);
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_PRIVILEGE_NOT_HELD);
        DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_ACCESS_DENIED);
        default:
          DVMESSAGE("AvSetMmThreadCharacteristics failed with"
              " unknow error %lu",dwLastError);
          break;
        }
      }
    }
#endif // ]

#if 0 // [
    if (eEngineAudioClient<=pEngine->eState) {
DVWRITELN("%s =========================================",EngineState(pEngine));
      if (EnginePlayingStart(pEngine)<0) {
        DMESSAGE("start playing");
        goto exit;
      }
DVWRITELN("%s =========================================",EngineState(pEngine));
    }
    else
#endif // ]
      EnginePlay(pEngine);
#if defined (YASAPI_THREAD_PLAYING) // [
    pThread->bPlaying=1;
#endif // ]
    DWRITELN(EngineState(pEngine));
    break;
  case eCommandFlush:
    ThreadKill(pThread,eEngineNull);
    EngineFlush(pEngine);
    DWRITELN(EngineState(pEngine));
    break;
  case eCommandKillDevice:
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
    ThreadKill(pThread,eEngineDevice);
#else // ] [
    ThreadKill(pThread,eEngineBase);
#endif // ]
    break;
#if defined (YASAPI_OPTION_DISCONNECT) // [
  case eCommandKillPauseDisconnected:
    ThreadKill(pThread,eEnginePauseDisconnected);
#if defined (YASAPI_THREAD_PLAYING) // [
    pThread->bPlaying=0;
#endif // ]
    break;
#else // ] [
  case eCommandKillPause:
    ThreadKill(pThread,eEnginePause);
    break;
#endif // ]
#if defined (YASAPI_OPTION_DISCONNECT) // [
  case eCommandKillPauseConnected:
#if defined (YASAPI_PAUSE_CONNECTED2) // [
    ThreadKill(pThread,eEnginePauseConnected2);
#else // ] [
    ThreadKill(pThread,eEnginePauseConnected);
#endif // ]
    break;
#endif // ]
  default:
    pEngine->thread.eCommand=eCommandNull;
    SetEvent(pEngine->thread.hConsumer);
    goto exit;
  }

  pEngine->thread.eCommand=eCommandNull;
  SetEvent(pEngine->thread.hConsumer);

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

static int ThreadPlay(Thread *pThread)
{
  int code=-1;
  Engine *pEngine=pThread->pEngine;
  Ring *pRing=&pEngine->ring;

  switch (pEngine->eState) {
  case eEngineEot:
#if defined (YASAPI_OPTION_DISCONNECT) // [
#if defined (YASAPI_PAUSE_CONNECTED2) // [
  case eEnginePauseConnected2:
#else // ] [
  case eEnginePauseConnected:
#endif // ]
#endif // ]
#if defined (YASAPI_OPTION_DISCONNECT) // [
  case eEnginePauseDisconnected:
#else // ] [
  case eEnginePause:
#endif // ]
    break;
  case eEngineDraining:
    if (0ul==pRing->lpVtbl->target.Written(pRing)) {
      ++pEngine->eState;
      DWRITELN(EngineState(pEngine));
      DASSERT(eEngineEot==pEngine->eState,exit);
      break;
    }

    // Intentional fall-through.
  default:
#if defined (YASAPI_THREAD_PLAYING) // [
    if (pThread->bPlaying&&eEngineStarted<=pEngine->eState) {
#else // ] [
    if (eEngineStarted<=pEngine->eState) {
#endif // ]
      if (EnginePlay(pEngine)<0) {
        DMESSAGE("playing");
        goto exit;
      }
#if ! defined (YASAPI_THREAD_PLAYING) // [
    }
#else // ] [
    }
#endif // ]

    break;
  }

  code=0;
exit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
DWORD WINAPI EngineThreadProc(LPVOID lpParameter)
{
  Thread thread;
  DWORD dwWait;

  ThreadCreate(&thread,lpParameter); // {

  for (;;) {
    dwWait=ThreadWait(&thread);

    switch (dwWait) {
    case WAIT_OBJECT_0+0ul:
      if (ThreadCommand(&thread)<0)
        goto exit;

      break;
    case WAIT_OBJECT_0+1ul:
    case WAIT_OBJECT_0+2ul:
      if (ThreadPlay(&thread)<0) {
        DMESSAGE("playing");
        goto exit;
      }

      break;
    default:
      DVMESSAGE("wait error: %lu",dwWait);
      goto exit;
    }
  }
exit:
  ThreadDestroy(&thread); // }
  return 0ul;
}
